// redFrik 060102
// http://www.pagat.com/vying/pokerrank.html

RedCard {
	var <value, <suit, <name;
	*new {|index| ^super.new.init(index)}		//index 0-51
	init {|argIndex|
		var index;
		if(argIndex<0 or:{argIndex>51}, {"expected index 0-51. truncated".warn});
		index= argIndex.clip(0, 51);
		value= index.div(4)+2;
		suit= switch(index%4,
			0, {"clubs"},
			1, {"diamonds"},
			2, {"hearts"},
			3, {"spades"}
		);
		name= suit++case
			{value==11} {$J}
			{value==12} {$Q}
			{value==13} {$K}
			{value==14} {$A}
			{true} {value}
	}
	printOn {|stream| stream<<this.name}
}
RedDeck {
	var <cards;
	*new {^super.new.init}
	init {this.shuffle}
	shuffle {^cards= (0..51).scramble.collect{|x| RedCard(x)}}
	drawCard {
		var card= cards.pop;
		if(card.isNil, {"out of cards.".warn});
		^card;
	}
	drawCards {|num| ^{this.drawCard}.dup(num)}
	cardsLeft {^cards.size}
}
RedHand {
	var <>cards;
	*new {|deck| ^super.new.init(deck)}
	init {|argDeck| cards= argDeck.drawCards(5).reject{|x| x.isNil}.sort{|x, y| x.value<y.value}}
	values {^cards.collect{|x| x.value}}
	suits {^cards.collect{|x| x.suit}}
}
RedPoker {
	compare {|...hands|
		var winner, winnerScore= -1, winnerRank;
		hands.do{|x, i|
			var hand, score, rank;
			hand= this.analyse(x);
			score= hand[0];
			if(score==winnerScore, {
				rank= hand[1];
				if((rank>winnerRank).any{|x| x}, {
					winner= i;
					winnerScore= score;
					winnerRank= rank;
				}, {
					if(rank==winnerRank, {
						winner= \tie;
					});
				});
			}, {
				if(score>winnerScore, {
					winner= i;
					winnerScore= score;
					winnerRank= hand[1];
				});
			});
		};
		^winner
	}
	analyse {|hand|
		var values, tempValues, count, flush, straight, royal, three, pairs;
		values= hand.values;
		count= values.asSet.asArray.collect{|x| values.occurencesOf(x)};
		flush= hand.suits.asSet.size==1;
		tempValues= values.wrap(1, 13).sort;
		straight= tempValues.drop(1).every{|x, i| x==(tempValues[i]+1)};
		royal= straight and:{values[0]>=10};
		three= count.includes(3);
		pairs= count.occurencesOf(2);
		^case
			{royal&&flush} {[
				9,
				#[14],
				\royalFlush
			]}
			{straight&&flush} {[
				8,
				if(values.sum==27, {	//ace counted as one. 2+3+4+5+13
					[values.removing(13).maxItem];
				}, {
					[values.maxItem]
				}),
				\straightFlush
			]}
			{count.includes(4)} {[
				7,
				[values.detect{|x| values.occurencesOf(x)==4}],
				\fourOfAKind
			]}
			{three&&(pairs==1)} {[
				6,
				[
					values.detect{|x| values.occurencesOf(x)==3},
					values.detect{|x| values.occurencesOf(x)==2}
				],
				\fullHouse
			]}
			{flush} {[
				5,
				values.sort{|x, y| x>y};
				\flush
			]}
			{straight} {[
				4,
				if(values.sum==27, {	//ace counted as one. 2+3+4+5+13
					[values.removing(13).maxItem];
				}, {
					[values.maxItem]
				}),
				\straight
			]}
			{three} {[
				3,
				[
					values.detect{|x| values.occurencesOf(x)==3},
					values.select{|x| values.occurencesOf(x)!=3}.sort{|x, y| x>y}
				].flat,
				\threeOfAKind
			]}
			{pairs==2} {[
				2,
				[
					values.select{|x| values.occurencesOf(x)==2}.sort{|x, y| x>y},
					values.detect{|x| values.occurencesOf(x)==1}
				].flat,
				\twoPairs
			]}
			{pairs==1} {[
				1,
				[
					values.detect{|x| values.occurencesOf(x)==2},
					values.select{|x| values.occurencesOf(x)==1}.sort{|x, y| x>y}
				].flat,
				\pair
			]}
			{true} {[
				0,
				values.sort{|x, y| x>y},
				\highCard
			]}
	}
}

/*
a= RedDeck.new
a.shuffle
a.cards
a.drawCard
a.cardsLeft
a.cards.count{|x| x.name.includes($1)}

b= RedCard(49)
b.value
b.suit
b.name

c= RedHand(a)
c.cards
d= RedHand(a)
d.cards

a= RedDeck.new
b= RedHand(a)
c= RedHand(a)
b.cards
c.cards
b.cards_([RedCard(0), RedCard(4), RedCard(8), RedCard(12), RedCard(16)]) //straightFlush
c.cards_([RedCard(5), RedCard(10), RedCard(13), RedCard(18), RedCard(23)]) //straight
b.cards_([RedCard(32), RedCard(37), RedCard(40), RedCard(45), RedCard(46)]) //pairs
b.cards_([RedCard(0), RedCard(1), RedCard(2), RedCard(3), RedCard(46)])
b.cards_([RedCard(46), RedCard(1), RedCard(1), RedCard(2), RedCard(3)])
b.cards_([RedCard(46), RedCard(47), RedCard(1), RedCard(2), RedCard(3)])
b.cards
d= RedPoker.new
d.analyse(b)
d.analyse(c)
d.compare(b, c)
*/
