// f.olofsson 041119-22, 041208, 041217
// 060916 rewrite from scratch

//need for looping?
//fix for the voice release issue?
//need for checking how many voices left?  or when a voice finished playing?

RedDiskInSampler {
	var <server, <>numFrames= 32768, <>overlaps= 2, keys;
	*initClass {
		StartUp.add{
			2.do{|i|								//change here for more channels than stereo
				SynthDef.writeOnce("redDiskInSampler-"++(i+1), {
					|i_out= 0, bufnum, amp= 0.7, attack= 0.01, sustain, release= 0.1, gate= 1|
					var src, env;
					src= DiskIn.ar((i+1), bufnum);
					env= EnvGen.kr(
						Env(#[0, 1, 1, 0], [attack, sustain, release], -4),
						gate,
						amp,
						0,
						1,
						2						//doneAction
					);
					Out.ar(i_out, src*env);
				}, [\ir]);
			}
		}
	}
	*new {|server|								//if server=nil it will use Server.default
		^super.new.initRedDiskInSampler(server);
	}
	initRedDiskInSampler {|s|
		keys= ();
		server= s ?? Server.default;
		CmdPeriod.add({this.prCmdStop});				//remove this too when .free
	}
	preload {|key, path, startFrame= 0|				//read file headers and allocate in advance
		var sf;
		if(server.serverRunning.not, {(this.class.asString++": server not running").error; this.halt});
		sf= SoundFile.new;
		if(sf.openRead(path).not, {(this.class.asString++": file not found_ "++path).error; sf.close; this.halt});
		keys.put(key, {|i|							//associate key with list of voice object
			var buffer;
			buffer= this.prAllocBuffer(numFrames, sf, path, startFrame);
			RedDiskInSamplerVoice(buffer, sf.numFrames/sf.sampleRate, path, startFrame);
		}.dup(overlaps));
		sf.close;
	}
	loadedKeys {^keys.keys}
	length {|key|
		var voices;
		if((voices= keys[key]).notNil, {^voices[0].length}, {^nil});
	}
	channels {|key|
		var voices;
		if((voices= keys[key]).notNil, {^voices[0].channels}, {^nil});
	}
	buffers {|key|
		var voices;
		if((voices= keys[key]).notNil, {^voices.collect{|x| x.buffer}}, {^nil});
	}
	voicesLeft {|key|
		var voices;
		if((voices= keys[key]).notNil, {^voices.count{|x| x.isPlaying==0}}, {^nil});
	}
	free {										//free synths, close files and free buffers
		keys.do{|voices|
			voices.do{|x|
				server.sendMsg("/n_free", x.nodeID);
				x.buffer.close;
				x.buffer.free;
			}
		};
		keys= ();
	}
	//play with finite duration - if sustain=nil then use file length
	play {|key, attack= 0, sustain, release= 0, amp= 0.7, out= 0, group|
		var voices, voice, grp, sus;
		voices= this.prFindVoices(key);
		voice= voices.detect{|x| x.isPlaying==0};		//find first voice ready to play
		if(voice.isNil, {(this.class.asString++": no free slots - increase overlaps").error; this.halt});
		grp= group ?? {server.defaultGroup};
		if(grp.isNumber.not, {grp= grp.nodeID});		//make it work with both IDs and objects
		sus= sustain ?? {(voice.length-attack-release).max(0)};
		voice.nodeID_(server.nextNodeID);
		voice.isPlaying_(2);
		this.prPlay(voice, sus, attack, release, grp, out, amp);
	}
	stop {|key, release= 0.4|
		var voices, voice;
		voices= this.prFindVoices(key);
		voice= voices.detect{|x| x.isPlaying==2};		//find first voice playing
		if(voice.isNil, {(this.class.asString++": all voices already stopped").error; this.halt});
		voice.isPlaying_(1);						//2 is playing, 1 stopped, 0 not playing
		server.sendMsg("/n_set", voice.nodeID, \gate, -1-release.min(voice.stopTime-Main.elapsedTime));
	}
	flush {|release= 0.4|							//cannot force close so voices lingers on
		keys.do{|voices|
			voices.do{|voice|
				if(voice.isPlaying==2, {
					server.sendMsg("/n_set", voice.nodeID, \gate, -1-release);
				});
			}
		}
	}
	//--private
	prFindVoices {|key|
		var voices;
		voices= keys[key];							//find all voices objects for this key
		if(voices.isNil, {(this.class.asString++": key not found").error; this.halt});
		^voices
	}
	prCmdStop {
		keys.do{|voices|
			voices.do{|voice|
				voice.buffer.close;
				voice.nodeID_(nil);
				voice.isPlaying_(0);
			}
		}
	}
	prAllocBuffer {|numFrames, sf, path, startFrame|
		^Buffer.alloc(server, numFrames, sf.numChannels)
	}
	prPlay {|voice, sus, attack, release, grp, out, amp|
		voice.buffer.cueSoundFile(voice.path, voice.startFrame, {
			//the following is done after cueSoundFile finished
			voice.stopTime_(Main.elapsedTime+sus+attack+release);
			SystemClock.schedAbs(voice.stopTime, {
				voice.buffer.close;
				voice.nodeID_(nil);
				voice.isPlaying_(0);
				nil;
			});
			["/s_new", "redDiskInSampler-"++voice.channels, voice.nodeID, 1, grp,				\i_out, out, \bufnum, voice.buffer.bufnum, \amp, amp,
				\attack, attack, \sustain, sus, \release, release
			]
		})
	}
}
RedDiskInSamplerGiga : RedDiskInSampler {
	//fast trigger version that re-cue files on .stop
	//note: this can only handle ca 244 files due to system limitations of how many files allwed to be kept open at the same time!
	prAllocBuffer {|numFrames, sf, path, startFrame|
		^Buffer.cueSoundFile(server, path, startFrame, sf.numChannels, numFrames)
	}
	prPlay {|voice, sus, attack, release, grp, out, amp|
		voice.stopTime_(Main.elapsedTime+sus+attack+release);
		SystemClock.schedAbs(voice.stopTime, {
			voice.buffer.close;
			voice.buffer.cueSoundFile(voice.path, voice.startFrame, {
				//the following is done after cueSoundFile finished
				voice.isPlaying_(0);
				voice.nodeID_(nil);
				nil;
			});
			nil;
		});
		server.sendMsg("/s_new", "redDiskInSampler-"++voice.channels, voice.nodeID, 1, grp,
			\i_out, out, \bufnum, voice.buffer.bufnum, \amp, amp,
			\attack, attack, \sustain, sus, \release, release
		)
	}
}
RedDiskInSamplerVoice {
	var <>isPlaying= 0, <>nodeID, <>stopTime, <buffer, <length, <path, <startFrame;
	*new {|buffer, length, path, startFrame|
		^super.new.initRedDiskInSamplerVoice(buffer, length, path, startFrame);
	}
	initRedDiskInSamplerVoice {|argBuf, argLen, argPath, argStart|
		buffer= argBuf;
		length= argLen;
		path= argPath;
		startFrame= argStart;
	}
	channels {^buffer.numChannels}
}

/*	-- debugcode --
s.boot;
s.stopAliveThread;
s.dumpOSC(1)
d= RedDiskInSampler.new(s);	//normal
d= RedDiskInSamplerGiga.new(s);	//normal
d.preload(\e, "sounds/bosen44100/56/E2.aif");
d.play(\e, 0, 0.11, 0.1);
d.play(\e, 0, 0.3, 0);
d.play(\e, 0, 0.1, 0);
d.play(\e, 0, nil, 0);
d.play(\e, 2.1, 0.3, 0);
d.stop(\e, 5)	//slight problem here
d.stop(\e, 0)	//slight problem here
d.flush
d.free
*/


/*	-- testcode --
s.boot;
d= RedDiskInSampler.new(s);	//normal
d= RedDiskInSamplerGiga.new(s);	//or fast trigger version
d.preload(\a, "sounds/bosen44100/56/A2.aif");
d.preload(\f, "sounds/bosen44100/40/F2.aif");
d.preload(\e, "sounds/bosen44100/56/E2.aif");
d.voicesLeft(\e)
d.play(\e, 0.2, 0.5, 1);	//key, attackTime, amp, outbus
d.stop(\e, 1.0);			//key, releaseTime
d.play(\e, 0.1, amp:1);
d.stop(\e, 2.1);
d.play(\e);
d.stop(\e);
d.play(\a, attack: 1);
d.play(\e, attack: 1);
d.stop(\e, release: 1);
d.stop(\a, release: 1);

d.play(\e);
d.play(\a);
d.play(\f);
d.play(\e);
d.play(\a);
d.play(\f);
d.flush(release: 0.2)

d.free;
*/


/*	-- more testcode --
s.boot;
~srcGrp= Group.head(s);
~efxGrp= Group.tail(s);
d= RedDiskInSampler.new(s);	//normal
d= RedDiskInSamplerGiga.new(s);	//or fast trigger version
d.preload(\e, "sounds/bosen44100/56/E2.aif");
e= SynthDef("ringmod", {Out.ar(0, In.ar(0, 2)*LPF.ar(BrownNoise.ar, 90))}).play(~efxGrp);
d.play(\e);
d.stop(\e);
d.play(\e, group: ~srcGrp);
d.stop(\e);
s.queryAllNodes;
d.free;
e.free;

i= Server.internal.boot;
d= RedDiskInSampler.new(i);	//normal
d= RedDiskInSamplerGiga.new(i);	//or fast trigger version
d.preload(\e, "sounds/bosen44100/56/E2.aif");
d.play(\e);
d.stop(\e);
d.free;
i.quit;
*/

