// /fO 050131
// keep all strings in a column equal length by padding with spaces (or other pad char)
// note: will convert everything to strings

RedExcel : Array2D {
	
	var <tabs, <pad= $ , <align= \left, lastLength= 0;
	
	init {|argRows, argCols|
		rows= argRows;
		cols= argCols;
		array= Array.fill(rows*cols, {""});
		tabs= Array.newClear(cols);
	}
	
	*fromString {|rows, cols, string, separator= $ |
		^this.fromArray(rows, cols, string.split(separator));
	}
	
	*fromStringsAsRows {|aarray, separator= $ |
		var rows, cols, array;
		rows= aarray.size;
		cols= 0;
		array= [];
		aarray.do{|i|
			i= i.split(separator);
			if(i.size>cols, {cols=i.size});
			array= array++i;
		};
		^this.fromArray(rows, cols, array);
	}
	
	with {|aarray|
		array= aarray.collect{|i| i.asString}.extend(rows*cols, "");
		this.prUpdateAll;
	}
	
	tabs_ {|tabArray|
		tabs= tabArray.clipExtend(cols);
		this.prUpdateAll;
	}
	
	pad_ {|char|
		array= array.collect{|i| i.tr(pad, char)};
		pad= char;
	}
	
	align_ {|symbol|
		align= symbol;
		this.prUpdateAll;
	}
	
	put {|row, col, str|
		var length;
		str= str.asString;
		array.put(row*cols+col, str);
		length= tabs[col] ?? {this.maxStringSize(col)};
		if(lastLength!=length, {
			this.extendStrings(col, length);
			lastLength= length;
		}, {
			this.prExtendString(row*cols+col, length);
		});
	}
	
	rowsAsString {|addRow= "\n", addCol= " "|
		var str;
		str= "";
		rows.do{|ri|
			str= str++this.rowAsString(ri, addCol)++addRow;
		};
		^str;
	}
	
	rowsAsArray {|addRow= "", addCol= " "|
		^Array.fill(rows, {|ri| this.rowAsString(ri, addCol)++addRow});
	}
	
	rowAsString {|row, addCol= " "|
		var str;
		str= "";
		cols.do{|ci|
			str= str++this.at(row, ci)++addCol;
		};
		^str;
	}
	
	maxStringSize {|col|	//remove pads and calculate length
		var maxSize= 0;
		rows.do{|ri|
			var size;
			size= this.at(ri, col).reject{|x| x==pad}.size;
			if(size>maxSize, {maxSize= size});
		};
		^maxSize;
	}
	
	maxLength {|col|		//calculate length including pads
		var maxLen= 0;
		rows.do{|ri|
			var str, size;
			str= this.at(ri, col);
			size= str.size;
			if(size>maxLen, {maxLen= size});
		};
		^maxLen;
	}
	
	extendStrings {|col, newLength|
		rows.do{|ri| this.prExtendString(ri*cols+col, newLength)};
	}
	
	//--private
	prExtendString {|index, newLength|
		var str, strSize;
		str= array.at(index) ? "";
		if(align==\left, {
			str= str.reject{|x| x==pad}.extend(newLength, pad);
		}, {
			if(align==\right, {
				str= str.reject{|x| x==pad}.reverse.extend(newLength, pad).reverse;
			}, {
				//else align \center
				str= str.reject{|x| x==pad};
				strSize= str.size;
				str= str.extend(newLength, pad);
				str= str.as(Array).rotate(((newLength-strSize)/2).asInteger).as(String);
			});
		});
		array.put(index, str);
	}
	
	prUpdateAll {
		cols.do{|ci|
			var length;
			length= tabs[ci] ?? {this.maxStringSize(ci)};
			this.extendStrings(ci, length);
		};
	}

}
